class ModelComponentNotFoundError(Exception):
    """Raised if a component is missing from the model."""

class ModelComponentError(Exception):
    """Raised if there is an issue with a component in the model."""
